"""Authenticate with services necessary for Halo Infinite API access."""

from aiohttp import ClientSession

from spnkr.auth import app, halo, oauth, player, xbox

XSTS_V3_XBOX_AUDIENCE = "http://xboxlive.com"
XSTS_V3_HALO_AUDIENCE = "https://prod.xsts.halowaypoint.com/"


async def authenticate_player(session: ClientSession, app: app.AzureApp) -> str:
    """Initial authentication of the `app` with Windows Live and Xbox Live.

    Request an OAuth token and return its refresh token. Save the refresh token
    for future use.

    Args:
        session: The session to use for requests.
        app: The Azure AD application to authenticate.

    Returns:
        The OAuth refresh token.
    """
    code = _get_authorization_code(app)
    token = await oauth.request_oauth_token(session, code, app)
    return token.refresh_token


def _get_authorization_code(app: app.AzureApp) -> str:
    """Prompt the user for an authorization code.

    Args:
        app: The Azure AD application to authenticate.

    Returns:
        The authorization code.
    """
    authorization_url = oauth.generate_authorization_url(app)
    print(authorization_url)
    print(
        "Navigate to the above URL and copy the 'code' parameter from the "
        "redirect URL query string in the address bar."
    )
    return input("Enter the code...").strip()


async def refresh_player_tokens(
    session: ClientSession, app: app.AzureApp, oauth_refresh_token: str
) -> player.AuthenticatedPlayer:
    """Refresh all tokens and return authenticated player information.

    Args:
        session: The session to use for requests.
        oauth_refresh_token: The OAuth refresh token.
        app: The Azure AD application to authenticate.

    Returns:
        The authenticated player information.
    """
    oauth_token = await oauth.refresh_oauth_token(session, oauth_refresh_token, app)
    user_token = await xbox.request_user_token(session, oauth_token.access_token)
    xsts_token = await xbox.request_xsts_token(
        session, user_token.token, XSTS_V3_XBOX_AUDIENCE
    )
    halo_xsts_token = await xbox.request_xsts_token(
        session, user_token.token, XSTS_V3_HALO_AUDIENCE
    )
    spartan_token = await halo.request_spartan_token(session, halo_xsts_token.token)
    clearance_token = await halo.request_clearance_token(session, spartan_token.token)
    return player.AuthenticatedPlayer(
        player_id=xsts_token.xuid,
        gamertag=xsts_token.gamertag,
        spartan_token=spartan_token,
        clearance_token=clearance_token,
        xbl_authorization_header_value=xsts_token.authorization_header_value,
    )



"""Read highlight event film chunks."""

import collections
import zlib
from typing import Any, Literal, NamedTuple

from bitstring import Bits

from spnkr.errors import FilmReadError
from spnkr.film.medals import MEDALS
from spnkr.models.stats import MatchStats, PlayerStats
from spnkr.xuid import unwrap_xuid

_MIN_XUID = int(2e15)
_MAX_XUID = int(3e15)
_XUID_BITS = 64
_MODE = 10
_DEATH = 20
_KILL = 50
_MEDAL_SORTING_WEIGHTS = {
    50,
    51,
    52,
    100,
    101,
    150,
    200,
    205,
    210,
    220,
    225,
    230,
    235,
    240,
    245,
    250,
}

EventType = Literal["kill", "death", "medal", "mode"]


class HighlightEvent(NamedTuple):
    """A kill, death, mode-related, or medal game event in a Halo Infinite match."""

    xuid: int
    """Xbox user ID of the player the event is tied to."""
    gamertag: str
    """Gamertag of the player the event is tied to."""
    type_hint: int
    """Byte value used to infer the event type."""
    is_medal: bool
    """Byte value used to infer the event type."""
    event_type: EventType
    """Inferred event type. One of 'kill', 'death', 'mode', or 'medal'."""
    time_ms: int
    """Timestamp of the event in milliseconds following the start of the match."""
    medal_value: int
    """Byte value encoding the medal type for medal event types."""
    medal_name: str | None
    """Name of the medal awarded for 'medal' event types. Otherwise `None`."""


def read(data: bytes, version: int):
    """Iterate events found in highlight event file content.

    Args:
        data: The gzip-compressed file data.
        version: The major version of the film, available from either the film metadata
            or the film header.

    Yields:
        Highlight events.
    """
    bits = Bits(bytes=zlib.decompress(data))
    for start, _ in _find_xuids(bits):
        yield _parse_event(bits, start, version)


def _find_xuids(bits: Bits):
    """Iterate XUID (start, value) tuples found in binary data."""
    # Example:
    # ------xuid------ --marker-
    # 2a5bf9c8f1010900 (2d|25)c0
    for marker_start in bits.findall(Bits(hex="c0")):
        xuid_end = marker_start - 8
        if bits[xuid_end:marker_start].hex not in ("2d", "25"):
            continue
        xuid_start = xuid_end - _XUID_BITS
        xuid_bits = bits[xuid_start:xuid_end]
        if len(xuid_bits) < _XUID_BITS:
            continue
        xuid = xuid_bits.uintle
        if _MIN_XUID < xuid < _MAX_XUID:
            yield (xuid_start, xuid)


def _infer_event_type(hint: int, is_medal: bool) -> EventType:
    """Infer the type of an event given information from a couple bytes."""
    if is_medal and hint in _MEDAL_SORTING_WEIGHTS:
        return "medal"
    if hint == _MODE:
        return "mode"
    if hint == _DEATH:
        return "death"
    if hint == _KILL:
        return "kill"
    raise FilmReadError(f"Unhandled event type args: {hint=}, {is_medal=}")


def _tokens(version: int) -> list[str]:
    """Determine unpacking strategy based on the film major version."""
    if version <= 38:
        return [
            "bytes:32",  # 16-character utf-16 gamertag (32 bytes)
            "pad:120",  # Skip (15 bytes)
            "uint:8",  # Type hint (1 byte)
            "uint:32",  # Timestamp in milliseconds (4 bytes)
            "pad:24",  # Skip (3 bytes)
            "uint:8",  # 1 if event is medal, otherwise 0 (1 byte)
            "pad:24",  # Skip (3 bytes)
            "uint:8",  # Medal type (1 byte)
        ]
    return [
        "pad:96",  # Skip (12 bytes)
        "bytes:32",  # 16-character utf-16 gamertag (32 bytes)
        "pad:24",  # Skip (3 bytes)
        "uint:8",  # Type hint (1 byte)
        "uint:32",  # Timestamp in milliseconds (4 bytes)
        "pad:24",  # Skip (3 bytes)
        "uint:8",  # 1 if event is medal, otherwise 0 (1 byte)
        "pad:24",  # Skip (3 bytes)
        "uint:8",  # Medal type (1 byte)
    ]


def _parse_event(bits: Bits, start: int, version: int) -> HighlightEvent:
    """Parse an event from `bits` starting at `start`."""
    selected = bits[start : start + 20_000]  # 20,000 bits just to capture the event
    # Grab the relevent 60 bytes of event data and unpack the attributes
    end = selected.find(Bits(hex="00002ee0"))[0]  # type: ignore
    event_bits = selected[end - (60 * 8) : end]
    values: list[Any] = event_bits.unpack(fmt=", ".join(_tokens(version)))
    gamertag_bytes, type_hint, time_ms, is_medal, medal_value = values
    event_type = _infer_event_type(type_hint, is_medal == 1)
    medal_name = MEDALS.get(medal_value) if event_type == "medal" else None
    return HighlightEvent(
        xuid=selected[:_XUID_BITS].uintle,
        gamertag=gamertag_bytes.decode("utf-16le").strip("\x00"),
        type_hint=type_hint,
        is_medal=is_medal == 1,
        event_type=event_type,
        time_ms=time_ms,
        medal_value=medal_value,
        medal_name=medal_name,
    )


class _EventCounts(NamedTuple):
    """Container for storing expected or resulting event counts."""

    kills: int
    deaths: int
    medals: int


def check(events: list[HighlightEvent], match_stats: MatchStats) -> list[str]:
    """Check that parsed highlight events line up with match stats from the API.

    For each player, check that the number of kills, deaths, and medals parsed
    into events match retrieved match stats from the API. Note that this check excludes
    "mode" event counts.

    It is expected that this check will fail for "minigame" or "firefight" game modes
    where players fight AI opponents. Killing these enemies do not trigger kill events,
    but they are accounted for in match stats. Errors include the game variant category
    in the message, which can justify why there are no kill events for a firefight match.

    There are occassionally kills and/or deaths accounted for in match stats that are
    not found in the highlight events chunk data, but these misses seem to be rare and
    are typically only one kill or death.

    Args:
        events: Parsed highlight events.
        match_stats: The match stats data from the Halo Infinite API.

    Returns:
        Errors, if any, caught while comparing `events` and `match_stats`.
    """
    out = []
    events_by_xuid = _group_events_by_xuid(events)
    for player in match_stats.players:
        if not player.is_human:
            continue
        expected = _get_player_stat_event_type_counts(player)
        player_events = events_by_xuid.get(unwrap_xuid(player.player_id), [])
        result = _get_highlight_event_type_counts(player_events)
        if result != expected:
            out.append(
                f"Expected {expected} for player {player.player_id}. Got {result}. "
                f"({match_stats.match_info.game_variant_category.name})"
            )
    return out


def _group_events_by_xuid(
    events: list[HighlightEvent],
) -> dict[int, list[HighlightEvent]]:
    """Group highlight `events` by their XUID."""
    out = collections.defaultdict(list)
    for event in events:
        out[event.xuid].append(event)
    return out


def _get_player_stat_event_type_counts(player: PlayerStats) -> _EventCounts:
    """Get expected event counts from player stats."""
    kills = deaths = medals = 0
    for player_team in player.player_team_stats:
        core = player_team.stats.core_stats
        kills += core.kills
        deaths += core.deaths
        medals += sum(m.count for m in core.medals)
    return _EventCounts(kills, deaths, medals)


def _get_highlight_event_type_counts(events: list[HighlightEvent]) -> _EventCounts:
    """Get event type counts for a list of `events`."""
    counter = collections.Counter(e.event_type for e in events)
    return _EventCounts(counter["kill"], counter["death"], counter["medal"])



"""Models for the "gamecms_hacs" authority."""

import datetime as dt

from pydantic import Field

from spnkr.models.base import CamelCaseModel, PascalCaseModel
from spnkr.models.refdata import MedalDifficulty, MedalNameId, MedalType
from spnkr.models.types import ReadOnlyDict


class TranslatableString(CamelCaseModel, frozen=True):
    """A string value accompanied by a dictionary of translations.

    Attributes:
        value: The string value.
        translations: A dictionary of language codes to translations for the string value.
    """

    value: str
    translations: ReadOnlyDict[str, str]


class Medal(CamelCaseModel, frozen=True):
    """Metadata for a single medal.

    Attributes:
        name_id: The medal's ID.
        name: The medal's name.
        description: The medal's description.
        sprite_index: The index of the medal's sprite in the sprite sheet.
        sorting_weight: The medal's sorting weight.
        difficulty: The medal's difficulty, such as "normal" or "legendary".
        type: The medal's type, such as "mode" or "proficiency".
        personal_score: The amount of personal score awarded by obtaining the medal.
    """

    name_id: MedalNameId
    name: TranslatableString
    description: TranslatableString
    sprite_index: int
    sorting_weight: int
    difficulty: MedalDifficulty = Field(alias="difficultyIndex")
    type: MedalType = Field(alias="typeIndex")
    personal_score: int


class SpriteSheet(CamelCaseModel, frozen=True):
    """Information about a sprite sheet that contains medal sprites.

    Attributes:
        path: The path to the sprite sheet, relative to the "gamecms_hacs" host.
        columns: The number of columns in the sprite sheet.
        size: The size of each sprite in the sprite sheet.
    """

    path: str
    columns: int
    size: int


class SpriteSheets(CamelCaseModel, frozen=True):
    """Sizes of sprite sheets that contain medal sprites.

    Attributes:
        small: A sprite sheet with small sprites.
        medium: A sprite sheet with medium sprites.
        extra_large: A sprite sheet with extra large sprites.
    """

    small: SpriteSheet
    medium: SpriteSheet
    extra_large: SpriteSheet = Field(alias="extra-large")


class MedalMetadata(CamelCaseModel, frozen=True):
    """Metadata for all medals.

    Attributes:
        difficulties: The list of possible difficulties.
        types: The list of possible types.
        sprites: Information about the medal sprite sheets that contain medal.
        medals: The list of available medals.
    """

    difficulties: tuple[str, ...]
    types: tuple[str, ...]
    sprites: SpriteSheets
    medals: tuple[Medal, ...]


class Date(PascalCaseModel, frozen=True):
    """A date object.

    Attributes:
        value: The date value.
    """

    value: dt.datetime = Field(alias="ISO8601Date")


class CsrSeason(PascalCaseModel, frozen=True):
    """A CSR season.

    Attributes:
        csr_season_file_path: The relative path to the CSR season file. The stem
            of the path is the CSR season's ID, e.g., "CsrSeason5-1".
        start_date: The UTC start date of the season.
        end_date: The UTC end date of the season.
    """

    csr_season_file_path: str
    start_date: Date
    end_date: Date


class CsrSeasonCalendar(PascalCaseModel, frozen=True):
    """A collection of past and current CSR seasons.

    Attributes:
        seasons: The list of CSR seasons.
    """

    seasons: tuple[CsrSeason, ...]


class Season(PascalCaseModel, frozen=True):
    """An promotional, reward-track operation contained within a CSR season.

    Attributes:
        csr_season_file_path: The relative path to the CSR season file. The stem
            of the path is the CSR season's ID, e.g., "CsrSeason5-1".
        operation_track_path: The relative path to the operation's reward track
            file.
        season_metadata: The relative path to the operation metadata file.
        start_date: The UTC start date of the season.
        end_date: The UTC end date of the season.
    """

    csr_season_file_path: str
    operation_track_path: str
    season_metadata: str
    start_date: Date
    end_date: Date


class Event(PascalCaseModel, frozen=True):
    """A promotional, reward-track event contained within a CSR season.

    Events were superseded by operations as of October 2023.

    Attributes:
        reward_track_path: The relative path to the event's reward track file.
        start_date: The UTC start date of the event.
        end_date: The UTC end date of the event.
    """

    reward_track_path: str
    start_date: Date
    end_date: Date


class CareerRank(PascalCaseModel, frozen=True):
    """Information related to career rank progression.

    Attributes:
        reward_track_path: The relative path to the career rank reward track
            file.
    """

    reward_track_path: str


class SeasonCalendar(PascalCaseModel, frozen=True):
    """A collection of past and current reward tracks.

    Attributes:
        seasons: The list of "operation" reward tracks.
        events: The list of "event" reward tracks. Events were superseded by
            operations as of October 2023.
        career_rank: Information about career rank progression.
    """

    seasons: tuple[Season, ...]
    events: tuple[Event, ...]
    career_rank: CareerRank


class InventoryReward(PascalCaseModel, frozen=True):
    """An inventory item reward, such as a weapon skin or armor piece.

    Attributes:
        inventory_item_path: The relative path to the inventory item file.
        amount: The quantity of the inventory item awarded.
        type: The type of inventory item, such as "ArmorCoating".
    """

    inventory_item_path: str
    amount: int
    type: str


class CurrencyReward(PascalCaseModel, frozen=True):
    """A currency reward, such as an XP boost.

    Attributes:
        currency_path: The relative path to the currency file.
        amount: The quantity of the currency awarded.
    """

    currency_path: str
    amount: int


class RankRewards(PascalCaseModel, frozen=True):
    """A collection of inventory and currency rewards earned at a rank.

    Attributes:
        inventory_rewards: The list of inventory item rewards.
        currency_rewards: The list of currency rewards.
    """

    inventory_rewards: tuple[InventoryReward, ...]
    currency_rewards: tuple[CurrencyReward, ...]


class TranslatableStringWithStatus(CamelCaseModel, frozen=True):
    """A string value accompanied by a dictionary of translations.

    Attributes:
        status: The status of the attribute.
        value: The string value.
        translations: A dictionary of language codes to translations for the
            string value.
    """

    status: str
    value: str
    translations: ReadOnlyDict[str, str]


class CareerRewardTrackRank(PascalCaseModel, frozen=True):
    """A rank in the career rank progression reward track.

    Title, subtitle, and grade can be combined for the rank's full name, e.g.,
    "Sergeant Bronze 2".

    Attributes:
        rank: The rank number.
        free_rewards: The rewards earned by all players that reach the rank.
        paid_rewards: The rewards earned by players that purchase the battle
            pass and reach the rank.
        xp_required_for_rank: The amount of XP required to reach the rank.
        rank_title: The rank's title, such as "Private" or "General".
        rank_sub_title: The rank's subtitle, such as "Bronze" or "Onyx".
        rank_tier: The rank's integer tier/grade - the smallest unit of rank
            progression.
        rank_icon: The relative path to the rank's icon file. Has a height of
            280px.
        rank_large_icon: The relative path to the rank's large icon file. This
            is the icon presented after a player acheives the rank. Has a height
            of 600px.
        rank_adornment_icon: The relative path to the rank's adornment icon
            file. Has a height of 196px.
        tier_type: The rank's tier type, such as "Bronze" or "Onyx".
        rank_grade: The rank's integer grade. Max is 3.
    """

    rank: int
    free_rewards: RankRewards
    paid_rewards: RankRewards
    xp_required_for_rank: int
    rank_title: TranslatableStringWithStatus
    rank_sub_title: TranslatableStringWithStatus
    rank_tier: TranslatableStringWithStatus
    rank_icon: str
    rank_large_icon: str
    rank_adornment_icon: str
    tier_type: str
    rank_grade: int


class CareerRewardTrack(PascalCaseModel, frozen=True):
    """The career rank progression reward track.

    Attributes:
        track_id: The reward track's ID.
        ranks: The list of ranks in the reward track.
        name: The reward track's name.
        description: The reward track's description.
        operation_number: The reward track's operation number.
        date_range: The reward track's date range.
        is_ritual: ...
        summary_image_path: The relative path to the reward track's summary
            image file.
        xp_per_rank: The amount of XP awarded per rank.
    """

    track_id: str
    ranks: tuple[CareerRewardTrackRank, ...]
    name: TranslatableStringWithStatus
    description: TranslatableStringWithStatus
    operation_number: int
    date_range: TranslatableStringWithStatus
    is_ritual: bool
    summary_image_path: str
    week_number: None
    xp_per_rank: int



"""Film-reading API"""

from uuid import UUID

from aiohttp import ClientResponseError

from spnkr.client import HaloInfiniteClient
from spnkr.errors import FilmReadError
from spnkr.film import highlight_events


async def read_highlight_events(client: HaloInfiniteClient, match_id: str | UUID):
    """Download and parse the highlight events chunk from a film asset for a given match.

    Only human (not bot) players have events available. Bot data is unlikely to be
    available in the highlight events chunk as they are not observable in theater
    mode.

    Team IDs for the players are not yet available, so it is necessary to join up
    to match stats data for the information.

    Mode-specific details are not available yet, such as a to distinguish a flag capture
    from a flag return.

    Killing AI opponents (such as in Firefight game modes) does not record a highlight
    event.

    There are infrequent cases where not all kill/death events are captured as
    expected. See `spnkr.film.highlight_events.check` for a method of checking
    outcomes against match stats.

    Args:
        client: A client for requesting film metadata and downloading film data.
        match_id: The UUID of the match to read film data for.

    Returns:
        A list of HighlightEvent instances extracted from the downloaded data.

    Raises:
        spnkr.errors.FilmReadError when film data can't be retrieved or read.
    """
    try:
        film_response = await client.discovery_ugc.get_film_by_match_id(match_id)
    except ClientResponseError as ex:
        raise FilmReadError(
            f"Error getting highlight events film metadata: {ex}"
        ) from ex
    film = await film_response.parse()
    url = film.highlight_events_url
    if url is None:
        raise FilmReadError("Film doesn't have a highlight events chunk")

    response = await client._session.get(url)
    try:
        response.raise_for_status()
    except ClientResponseError as ex:
        raise FilmReadError(
            f"Error downloading highlight events film chunk: {ex}"
        ) from ex

    version = film.custom_data.film_major_version
    return list(highlight_events.read(await response.read(), version))



"""Stats data services."""

import warnings
from typing import Literal
from uuid import UUID

from spnkr.models.refdata import GameplayInteraction, GameVariantCategory
from spnkr.models.stats import (
    MatchCount,
    MatchHistory,
    MatchStats,
    ServiceRecord,
)
from spnkr.responses import JsonResponse
from spnkr.services.base import BaseService
from spnkr.xuid import wrap_xuid_or_gamertag

_HOST = "https://halostats.svc.halowaypoint.com:443"
_VALID_SERVICE_RECORD_FILTER_SETS = [
    {"season_id"},
    {"season_id", "game_variant_category"},
    {"season_id", "game_variant_category", "playlist_asset_id"},
    {"season_id", "game_variant_category", "is_ranked"},
    {"season_id", "playlist_asset_id"},
    {"season_id", "gameplay_interaction"},
    {"game_variant_category"},
    {"game_variant_category", "is_ranked"},
]


class StatsService(BaseService):
    """Stats data services."""

    async def get_match_count(self, player: str | int) -> JsonResponse[MatchCount]:
        """Get match counts across different game experiences for a player.

        The counts returned are for custom matches, matchmade matches, local
        matches, and total matches.

        Args:
            player: Xbox Live ID or gamertag of the player to get counts for.
                Examples of valid inputs include "xuid(1234567890123456)",
                "1234567890123456", 1234567890123456, and "MyGamertag".

        Returns:
            The match counts.
        """
        xuid_or_gamertag = wrap_xuid_or_gamertag(player)
        url = f"{_HOST}/hi/players/{xuid_or_gamertag}/matches/count"
        resp = await self._get(url)
        return JsonResponse(resp, lambda data: MatchCount(**data))

    async def get_service_record(
        self,
        player: str | int,
        match_type: Literal["matchmade", "custom", "local"] = "matchmade",
        season_id: str | None = None,
        game_variant_category: GameVariantCategory | int | None = None,
        is_ranked: bool | None = None,
        playlist_asset_id: str | UUID | None = None,
        gameplay_interaction: GameplayInteraction | None = None,
    ) -> JsonResponse[ServiceRecord]:
        """Get a service record for a player. Summarizes player stats.

        Note that filters (`season_id`, `game_variant_category`, `is_ranked`,
        and `playlist_asset_id`) are only applicable to "matchmade"
        `match_type`. A warning is issued and the filters are ignored if they
        are provided for a non-matchmade `match_type`.

        Filters must be combined appropriately. The following are valid:

        - `season_id`
        - `season_id`, `game_variant_category`
        - `season_id`, `game_variant_category`, `playlist_asset_id`
        - `season_id`, `game_variant_category`, `is_ranked`
        - `season_id`, `playlist_asset_id`
        - `season_id`, `gameplay_interaction`
        - `game_variant_category`
        - `game_variant_category`, `is_ranked`

        To collect possible values for the filter arguments, look at the
        "subqueries" attribute of an unfiltered service record response.

        Args:
            player: Xbox Live ID or gamertag of the player to get counts for.
                Examples of valid inputs include "xuid(1234567890123456)",
                "1234567890123456", 1234567890123456, and "MyGamertag".
            match_type: The type of games to include in the service record.
                One of "matchmade", "custom", or "local".
            season_id: The season ID to get service record for. Optional.
            game_variant_category: The game variant category to filter service
                record data. See `spnkr.parsers.refdata.GameVariantCategory` for
                human-readable values. Optional.
            is_ranked: Filter for ranked or unranked games. Optional.
            playlist_asset_id: Filter for a specific playlist with its asset ID.
                Optional.
            gameplay_interaction: Filter for a specific gameplay interaction
                (e.g., PVP).

        Returns:
            The service record for the player with the given filters.

        Raises:
            ValueError: If `match_type` is not one of "matchmade", "custom", or
                "local".
            ValueError: If filter arguments are inappropriately combined.
        """
        if match_type.lower() not in ("matchmade", "custom", "local"):
            raise ValueError(f"Invalid match type: {match_type}")
        xuid_or_gamertag = wrap_xuid_or_gamertag(player)
        endpoint = f"/hi/players/{xuid_or_gamertag}/{match_type}/servicerecord"
        url = f"{_HOST}{endpoint}"
        filters = {
            "season_id": season_id,
            "game_variant_category": game_variant_category,
            "is_ranked": is_ranked,
            "playlist_asset_id": playlist_asset_id,
            "gameplay_interaction": gameplay_interaction,
        }
        filters = {k: v for k, v in filters.items() if v is not None}
        if match_type.lower() != "matchmade" and filters:
            warnings.warn(
                "Service record filters are only applicable to matchmade games."
            )
            filters = {}
        if filters and set(filters) not in _VALID_SERVICE_RECORD_FILTER_SETS:
            valid = "\n".join(str(s) for s in _VALID_SERVICE_RECORD_FILTER_SETS)
            raise ValueError(
                f"Invalid filter combination: {filters}. Options:\n{valid}"
            )
        params = {k.replace("_", ""): str(v) for k, v in filters.items()}
        resp = await self._get(url, params=params)
        return JsonResponse(resp, lambda data: ServiceRecord(**data))

    async def get_match_history(
        self,
        player: str | int,
        start: int = 0,
        count: int = 25,
        match_type: Literal["all", "matchmaking", "custom", "local"] = "all",
    ) -> JsonResponse[MatchHistory]:
        """Request a batch of matches from a player's match history.

        Args:
            player: Xbox Live ID or gamertag of the player to get counts for.
                Examples of valid inputs include "xuid(1234567890123456)",
                "1234567890123456", 1234567890123456, and "MyGamertag".
            start: Index of the first match to request, starting at 0.
            count: The number of matches to request. Maximum number of results
                returned is 25.
            match_type: The type of matches to return. One of "all",
                "matchmaking", "custom", or "local".

        Returns:
            The requested match history "page" of results.
        """
        xuid_or_gamertag = wrap_xuid_or_gamertag(player)
        url = f"{_HOST}/hi/players/{xuid_or_gamertag}/matches"
        params = {"start": start, "count": count, "type": match_type}
        resp = await self._get(url, params=params)
        return JsonResponse(resp, lambda data: MatchHistory(**data))

    async def get_match_stats(self, match_id: str | UUID) -> JsonResponse[MatchStats]:
        """Request match details using the Halo Infinite match GUID.

        Args:
            match_id: Halo Infinite GUID identifying the match.

        Returns:
            The match details.
        """
        url = f"{_HOST}/hi/matches/{match_id}/stats"
        resp = await self._get(url)
        return JsonResponse(resp, lambda data: MatchStats(**data))



"""Information about an authenticated player."""

import datetime as dt
from dataclasses import dataclass

from spnkr.auth.halo import ClearanceToken, SpartanToken


@dataclass(frozen=True)
class AuthenticatedPlayer:
    """Information about an authenticated player.

    Attributes:
        player_id: Xbox Live ID of the player.
        gamertag: The player's gamertag.
        spartan_token: An expirable Spartan token required for all API
            endpoints.
        clearance_token: A clearance token required for some API endpoints.
        xbl_authorization_header_value: The value for the authorization header
            passed when making requests to the Xbox Live API.
    """

    player_id: str
    gamertag: str
    spartan_token: SpartanToken
    clearance_token: ClearanceToken
    xbl_authorization_header_value: str

    @property
    def is_valid(self) -> bool:
        """Whether the player's Spartan token is still valid."""
        return self.spartan_token.expires_at > dt.datetime.now(dt.timezone.utc)



"""Enumerated data types used by the Halo Infinite API."""

import logging
from enum import Enum, IntEnum, StrEnum

logger = logging.getLogger(__name__)


class AssetHome(IntEnum):
    """Source of an asset."""

    UNKNOWN = 0
    STUDIO = 1
    PLAYER = 2


class AssetKind(IntEnum):
    """Types of assets used by Halo Infinite."""

    UNKNOWN = 0
    FILM = 1
    MAP = 2
    PLAYLIST = 3
    PREFAB = 4
    TEST_ASSET = 5
    UGC_GAME_VARIANT = 6
    MAP_MODE_PAIR = 7
    PROJECT = 8
    MANIFEST = 9
    ENGINE_GAME_VARIANT = 10


class BotDifficulty(IntEnum):
    """Bot difficulty levels as encoded in stats responses."""

    UNKNOWN = -1
    MARINE = 1
    ODST = 2
    SPARTAN = 3
    RECRUIT = 4
    ADAPTIVE = 5


class CloneBehavior(IntEnum):
    """Permission levels for cloning assets."""

    DEFAULT = 0
    TEMPLATE = 1
    PROHIBITED = 2


class FilmChunkType(IntEnum):
    """Types of saved film chunks."""

    NONE = 0
    FILM_HEADER = 1
    REPLICATION_DATA = 2
    HIGHLIGHT_EVENTS = 3


class FilmStatus(IntEnum):
    """Status of a saved film."""

    UNKNOWN = 0
    COMPLETE = 1
    ONGOING = 2
    RECOVERED = 3


class GameplayInteraction(IntEnum):
    """Types of enemies interacted with in-game."""

    # Retrieved from JavaScript files in Waypoint network traffic.
    UNKNOWN = -1
    NONE = 0
    PVP = 1
    PVE = 2
    PVPVE = 3


class GameVariantCategory(IntEnum):
    """Game mode category enumeration."""

    UNKNOWN = -1
    NONE = 0
    CAMPAIGN = 1
    FORGE = 2
    ACADEMY = 3
    ACADEMY_TUTORIAL = 4
    ACADEMY_PRACTICE = 5
    MULTIPLAYER_SLAYER = 6
    MULTIPLAYER_ATTRITION = 7
    MULTIPLAYER_ELIMINATION = 8
    MULTIPLAYER_FIESTA = 9
    MULTIPLAYER_SWAT = 10
    MULTIPLAYER_STRONGHOLDS = 11
    MULTIPLAYER_BASTION = 12
    MULTIPLAYER_KING_OF_THE_HILL = 13
    MULTIPLAYER_TOTAL_CONTROL = 14
    MULTIPLAYER_CTF = 15
    MULTIPLAYER_ASSAULT = 16
    MULTIPLAYER_EXTRACTION = 17
    MULTIPLAYER_ODDBALL = 18
    MULTIPLAYER_STOCKPILE = 19
    MULTIPLAYER_JUGGERNAUT = 20
    MULTIPLAYER_REGICIDE = 21
    MULTIPLAYER_INFECTION = 22
    MULTIPLAYER_ESCORT = 23
    MULTIPLAYER_GUN_GAME = 24
    MULTIPLAYER_GRIFBALL = 25
    MULTIPLAYER_RACE = 26
    MULTIPLAYER_PROTOTYPE = 27
    TEST = 28
    TEST_ACADEMY = 29
    TEST_AUDIO = 30
    TEST_CAMPAIGN = 31
    TEST_ENGINE = 32
    TEST_FORGE = 33
    TEST_GRAPHICS = 34
    TEST_MULTIPLAYER = 35
    TEST_SANDBOX = 36
    ACADEMY_TRAINING = 37
    ACADEMY_WEAPON_DRILL = 38
    MULTIPLAYER_LAND_GRAB = 39
    MULTIPLAYER_MINIGAME = 41
    MULTIPLAYER_FIREFIGHT = 42


class InspectionResult(IntEnum):
    """Related to readiness of user-generated content?"""

    UNKNOWN = 0
    TOLERABLE = 5
    STUDIO_CONTENT = 50


class LifecycleMode(IntEnum):
    """Match origin/hosting method enumeration."""

    CUSTOM = 1
    MATCHMADE = 3
    CUSTOM_LAN = 7


class MedalNameId(IntEnum):
    """Medal award enumeration."""

    def __new__(cls, value: int, display_name: str):
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj._display_name = display_name  # type: ignore
        return obj

    @property
    def display_name(self) -> str:
        """Get the name of the medal as displayed in-game."""
        return self._display_name  # type: ignore

    THREE_SIXTY = 1427176344, "360"
    ACE = 521420212, "Ace"
    ACHILLES_SPINE = 3217141618, "Achilles Spine"
    ACTION_HERO = 2976102155, "Action Hero"
    ALL_THAT_JUICE = 3528500956, "All That Juice"
    ALWAYS_ROTATING = 1472686630, "Always Rotating"
    APOCALYPSE = 3653884673, "Apocalypse"
    ASSURED_DESTRUCTION = 1585298941, "Assured Destruction"
    AUTOPILOT_ENGAGED = 1623236079, "Autopilot Engaged"
    BACK_SMACK = 548533137, "Back Smack"
    BALLISTA = 4215552487, "Ballista"
    BANK_SHOT = 2414983178, "Bank Shot"
    BIG_DEAL = 2125906504, "Big Deal"
    BLIGHT = 88914608, "Blight"
    BLIND_FIRE = 4007438389, "Blind Fire"
    BODYGUARD = 555849395, "Bodyguard"
    BOMBER = 1146876011, "Bomber"
    BOOGEYMAN = 1720896992, "Boogeyman"
    BOOM_BLOCK = 524758914, "Boom Block"
    BOXER = 269174970, "Boxer"
    BREACHER = 2750622016, "Breacher"
    BULLTRUE = 3114137341, "Bulltrue"
    CALL_BLOCKED = 2964157454, "Call Blocked"
    CHAIN_REACTION = 1969067783, "Chain Reaction"
    CLASH_OF_KINGS = 1053114074, "Clash of Kings"
    CLEANSING = 1765213446, "Cleansing"
    CLEAR_RECEPTION = 394349536, "Clear Reception"
    CLOCK_STOP = 3630529364, "Clock Stop"
    CLUSTER_LUCK = 3905838030, "Cluster Luck"
    COMBAT_EVOLVED = 641726424, "Combat Evolved"
    CONTRACT_KILLER = 590706932, "Contract Killer"
    COUNTER_SNIPE = 1477806194, "Counter-snipe"
    CULLING = 1025827095, "Culling"
    DEADLY_CATCH = 2396845048, "Deadly Catch"
    DEATH_CABBIE = 2848470465, "Death Cabbie"
    DEATH_RACE = 677323068, "Death Race"
    DEEP_BALL = 1254180082, "Deep Ball"
    DEMON = 2875941471, "Demon"
    DISEASE = 1155542859, "Disease"
    DIVINE_INTERVENTION = 2164872967, "Divine Intervention"
    DOGFIGHT = 1229018603, "Dogfight"
    DOUBLE_KILL = 622331684, "Double Kill"
    DRIVEBY = 197913196, "Driveby"
    DRIVER = 3027762381, "Driver"
    DRIVING_SPREE = 3169118333, "Driving Spree"
    DUELIST = 4247875860, "Duelist"
    EXTERMINATION = 4100966367, "Extermination"
    FAST_BREAK = 4014259917, "Fast Break"
    FAST_LANK = 3945864962, "Fast Lane"
    FASTBALL = 1211820913, "Fastball"
    FIRE_AND_FORGET = 988255960, "Fire & Forget"
    FLAG_JOUST = 976049027, "Flag Joust"
    FLAWLESS_VICTORY = 1680000231, "Flawless Victory"
    FLYIN_HIGH = 3739610597, "Flyin' High"
    FROM_THE_GRAVE = 2625820422, "From the Grave"
    FROM_THE_VOID = 3588869844, "From the Void"
    FUMBLE = 3732790338, "Fumble"
    GOAL_LINE_STAND = 3227840152, "Goal Line Stand"
    GRAND_SLAM = 1646928910, "Grand Slam"
    GRAPPLE_JACK = 690125105, "Grapple-jack"
    GREAT_JOURNEY = 1376646881, "Great Journey"
    GRENADIER = 2648272972, "Grenadier"
    GRIM_REAPER = 2567026752, "Grim Reaper"
    GUARDIAN_ANGEL = 3334154676, "Guardian Angel"
    GUNNER = 3783455472, "Gunner"
    GUNSLINGER = 1172766553, "Gunslinger"
    HAIL_MARY = 3934547153, "Hail Mary"
    HANG_UP = 4285712605, "Hang Up"
    HANG_TIME = 1325926691, "Hang Time"
    HARPOON = 2418616582, "Harpoon"
    HEAVY = 4086138034, "Heavy"
    HELLS_JANITOR = 217730222, "Hell's Janitor"
    HIGH_VALUE_TARGET = 3041030832, "High Value Target"
    HILL_GUARDIAN = 580478179, "Hill Guardian"
    HOLD_THIS = 175594566, "Hold This"
    IMMORTAL = 3120600565, "Immortal"
    IMMORTAL_CHAUFFEUR = 1739996188, "Immortal Chauffeur"
    INTERCEPTION = 2362950720, "Interception"
    INTERLINKED = 651256911, "Interlinked"
    KILLAMANJARO = 3835606176, "Killamanjaro"
    KILLING_FRENZY = 4261842076, "Killing Frenzy"
    KILLING_SPREE = 2780740615, "Killing Spree"
    KILLIONAIRE = 3233051772, "Killionaire"
    KILLJOY = 3233952928, "Killjoy"
    KILLPOCALYPSE = 3352648716, "Killpocalypse"
    KILLTACULAR = 2137071619, "Killtacular"
    KILLTASTROPHE = 2242633421, "Killtastrophe"
    KILLTROCITY = 1430343434, "Killtrocity"
    KONG = 3546244406, "Kong"
    LAST_SHOT = 3091261182, "Last Shot"
    LATE_BOOMER = 1334138090, "Late Boomer"
    LAWNMOWER = 3475540930, "Lawnmower"
    LONE_WOLF = 2623698509, "Lone Wolf"
    MARKSMAN = 2758320809, "Marksman"
    MEGANAUT = 2005352812, "Meganaut"
    MIND_THE_GAP = 1880789493, "Mind the Gap"
    MONOPOLY = 1090931685, "Monopoly"
    MOUNT_UP = 1065136443, "Mount Up"
    MOUNTED_AND_LOADED = 1331361851, "Mounted & Loaded"
    NADE_SHOT = 265478668, "Nade Shot"
    NECROMANCER = 3011158621, "Necromancer"
    NIGHTMARE = 710323196, "Nightmare"
    NINJA = 3085856613, "Ninja"
    NO_SCOPE = 2602963073, "No Scope"
    NUCLEAR_FOOTBALL = 2253222811, "Nuclear Football"
    ODINS_RAVEN = 87172902, "Odin's Raven"
    OFF_THE_RACK = 1283796619, "Off the Rack"
    OVERKILL = 835814121, "Overkill"
    PANCAKE = 3876426273, "Pancake"
    PARTYS_OVER = 3583966655, "Party's Over"
    PERFECT = 1512363953, "Perfect"
    PERFECTION = 865763896, "Perfection"
    PESTILENCE = 1719203329, "Pestilence"
    PILOT = 2593226288, "Pilot"
    PINEAPPLE_EXPRESS = 2019283350, "Pineapple Express"
    PLAGUE = 3786134933, "Plague"
    POWER_OUTAGE = 629165579, "Power Outage"
    PULL = 4132863117, "Pull"
    PURGE = 3467301935, "Purge"
    PURIFICATION = 496411737, "Purification"
    QUICK_DRAW = 2861418269, "Quick Draw"
    QUIGLEY = 1312042926, "Quigley"
    RAMMING_SPEED = 1298835518, "Ramming Speed"
    RAMPAGE = 1486797009, "Rampage"
    RECLAIMER = 1445036152, "Reclaimer"
    REMOTE_DETONATION = 3160646854, "Remote Detonation"
    RETURN_TO_SENDER = 3059799290, "Return to Sender"
    REVERSAL = 2123530881, "Reversal"
    RIDESHARE = 656245292, "Rideshare"
    RIFLEMAN = 2852571933, "Rifleman"
    RUNNING_RIOT = 418532952, "Running Riot"
    SABOTEUR = 20397755, "Saboteur"
    SCATTERGUNNER = 3347922939, "Scattergunner"
    SCOURGE = 3520382976, "Scourge"
    SECURE_LINE = 2426456555, "Secure Line"
    SHARPSHOOTER = 4277328263, "Sharpshooter"
    SHOT_CALLER = 1169571763, "Shot Caller"
    SIGNAL_BLOCK = 3931425309, "Signal Block"
    SKYJACK = 731054446, "Skyjack"
    SNEAK_KING = 670606868, "Sneak King"
    SNIPE = 4229934157, "Snipe"
    SOLE_SURVIVOR = 2717755703, "Sole Survivor"
    SPECIAL_DELIVERY = 275666139, "Special Delivery"
    SPLATTER = 221693153, "Splatter"
    SPOTTER = 2477555653, "Spotter"
    STEAKTACULAR = 1169390319, "Steaktacular"
    STICK = 3655682764, "Stick"
    STOPPED_SHORT = 3488248720, "Stopped Short"
    STRAIGHT_BALLING = 781229683, "Straight Balling"
    STREET_SWEEPER = 2967011722, "Street Sweeper"
    TAG_AND_BAG = 1841872491, "Tag & Bag"
    TANKER = 2278023431, "Tanker"
    THE_SICKNESS = 17866865, "The Sickness"
    TREASURE_HUNTER = 1685043466, "Treasure Hunter"
    TRIPLE_KILL = 2063152177, "Triple Kill"
    UNDEAD_HUNTER = 1447057920, "Undead Hunter"
    UNTAINTED = 1064731598, "Untainted"
    WARRIOR = 1210678802, "Warrior"
    WATCH_THE_THRONE = 1210968206, "Watch the Throne"
    WHEELMAN = 2926348688, "Wheelman"
    WHIPLASH = 1734214473, "Whiplash"
    WINDSHIELD_WIPER = 2827657131, "Windshield Wiper"
    WINGMAN = 1284032216, "Wingman"
    YARD_SALE = 1176569867, "Yard Sale"
    ZOMBIE_SLAYER = 557309779, "Zombie Slayer"
    ZONE_GUARDIAN = 420808268, "Zone Guardian"


class MedalDifficulty(IntEnum):
    """Difficulty of medals obtainable in matchmaking."""

    # Values line up to indices in the medal metadata response content.
    NORMAL = 0
    HEROIC = 1
    LEGENDARY = 2
    MYTHIC = 3


class MedalType(IntEnum):
    """Types of medals obtainable in matchmaking."""

    # Values line up to indices in the medal metadata response content.
    SPREE = 0
    MODE = 1
    MULTIKILL = 2
    PROFICIENCY = 3
    SKILL = 4
    STYLE = 5


class Outcome(IntEnum):
    """Match outcome options."""

    UNKNOWN = -1
    NONE = 0
    TIE = 1
    WIN = 2
    LOSS = 3
    DID_NOT_FINISH = 4
    DID_NOT_START = 5


class PersonalScoreNameId(IntEnum):
    """Personal score award enumeration."""

    def __new__(cls, value: int, display_name: str):
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj._display_name = display_name  # type: ignore
        return obj

    @property
    def display_name(self) -> str:
        """Get the name of the personal score award as displayed in-game."""
        return self._display_name  # type: ignore

    BALL_CONTROL = 454168309, "Ball Control"  # 50
    BALL_TAKEN = 204144695, "Ball Taken"  # 10
    BETRAYED_PLAYER = 911992497, "Betrayed Player"  # -100
    CARRIER_KILLED = 4128329646, "Carrier Killed"  # 10, Stockpile
    CARRIER_STOPPED = 746397417, "Carrier Stopped"  # 25, Oddball
    COLLECTED_BONUS_XP = 522435689, "Collected Bonus XP"  # 300, Last Spartan Standing
    CONVERSION_DENIED = 4247243561, "Conversion Denied"  # 25
    CUSTOM = 4294967295, "Custom"  # Variable, scripted in a node graph
    DESTROYED_BANSHEE = 597066859, "Destroyed Banshee"  # 50
    DESTROYED_CHOPPER = 3472794399, "Destroyed Chopper"  # 50
    DESTROYED_FALCON = 395875864, "Destroyed Falcon"  # 75
    DESTROYED_GHOST = 4254982885, "Destroyed Ghost"  # 50
    DESTROYED_GUNGOOSE = 2107631925, "Destroyed Gungoose"  # 25
    DESTROYED_MONGOOSE = 1416267372, "Destroyed Mongoose"  # 25
    DESTROYED_PHANTOM = 2742351765, "Destroyed Phantom"  # 100
    DESTROYED_RAZORBACK = 1661163286, "Destroyed Razorback"  # 50
    DESTROYED_ROCKET_WARTHOG = 2008690931, "Destroyed Rocket Warthog"  # 50
    DESTROYED_SCORPION = 3454330054, "Destroyed Scorpion"  # 100
    DESTROYED_WARTHOG = 3107879375, "Destroyed Warthog"  # 50
    DESTROYED_WASP = 2106274556, "Destroyed Wasp"  # 50
    DESTROYED_WRAITH = 3243589708, "Destroyed Wraith"  # 100
    DRIVER_ASSIST = 963594075, "Driver Assist"  # 50
    EMP_ASSIST = 221060588, "EMP Assist"  # 50
    ELIMINATED_PLAYER = 2408971842, "Eliminated Player"  # 200
    EXTRACTION_COMPLETED = 4130011565, "Extraction Completed"  # 200
    EXTRACTION_CONVERTED = 1117301492, "Extraction Converted"  # 50
    EXTRACTION_DENIED = 1552628741, "Extraction Denied"  # 25
    EXTRACTION_INITIATED = 1825517751, "Extraction Initiated"  # 50
    FLAG_CAPTURE_ASSIST = 555570945, "Flag Capture Assist"  # 100
    FLAG_CAPTURED = 601966503, "Flag Captured"  # 300
    FLAG_RETURNED = 22113181, "Flag Returned"  # 25
    FLAG_STOLEN = 3002710045, "Flag Stolen"  # 25
    FLAG_TAKEN = 2387185397, "Flag Taken"  # 10
    HACKED_TERMINAL = 665081740, "Hacked Terminal"  # 100
    HIJACKED_BANSHEE = 3150095814, "Hijacked Banshee"  # 25
    HIJACKED_CHOPPER = 1059880024, "Hijacked Chopper"  # 25
    HIJACKED_FALCON = 586857799, "Hijacked Falcon"  # 25
    HIJACKED_GHOST = 1614285349, "Hijacked Ghost"  # 25
    HIJACKED_GUNGOOSE = 4186766732, "Hijacked Gungoose"  # 25
    HIJACKED_MONGOOSE = 2191528998, "Hijacked Mongoose"  # 25
    HIJACKED_RAZORBACK = 2848565291, "Hijacked Razorback"  # 25
    HIJACKED_ROCKET_WARTHOG = 4294405210, "Hijacked Rocket Warthog"  # 25
    HIJACKED_WARTHOG = 1834653062, "Hijacked Warthog"  # 25
    HIJACKED_WASP = 674964649, "Hijacked Wasp"  # 25
    HILL_CONTROL = 340198991, "Hill Control"  # 25
    HILL_SCORED = 1032565232, "Hill Scored"  # 100
    KILL_ASSIST = 638246808, "Kill Assist"  # 50
    KILLED_PLAYER = 1024030246, "Killed Player"  # 100
    MARK_ASSIST = 152718958, "Mark Assist"  # 10
    POWER_SEED_SECURED = 2188620691, "Power Seed Secured"  # 100
    POWER_SEED_STOLEN = 3996338664, "Power Seed Stolen"  # 50
    REVIVE_DENIED = 2130209372, "Revive Denied"  # 25
    REVIVED_PLAYER = 3428202435, "Revived Player"  # 100
    RUNNER_STOPPED = 316828380, "Runner Stopped"  # 25, CTF
    SELF_DESTRUCTION = 249491819, "Self-destruction"  # -100
    SENSOR_ASSIST = 1267013266, "Sensor Assist"  # 10
    STOCKPILE_SCORED = 2801241965, "Stockpile Scored"  # 150
    ZONE_CAPTURED_50 = 3507884073, "Zone Captured"  # 50
    ZONE_CAPTURED_75 = 4026987576, "Zone Captured"  # 75
    ZONE_CAPTURED_100 = 757037588, "Zone Captured"  # 100
    ZONE_SECURED = 709346128, "Zone Secured"  # 25


class PlayerType(IntEnum):
    """Types of players."""

    UNKNOWN = -1
    HUMAN = 1
    BOT = 2


class PlaylistBotDifficulty(IntEnum):
    """Bot difficulty options for matchmaking playlists."""

    AUTOMATIC = 0
    RECRUIT = 1
    MARINE = 2
    ODST = 3
    SPARTAN = 4


class PlaylistDeviceInput(IntEnum):
    """Device input options for matchmaking playlists."""

    UNKNOWN = 0
    CONTROLLER = 1
    MOUSE_KEYBOARD = 2


class PlaylistEntrySelectionStrategy(IntEnum):
    """Selection strategies for playlist entries."""

    WEIGHTED = 0
    NO_REPEAT = 1


class PlaylistExperience(IntEnum):
    """Type of experience provided by a matchmaking playlist."""

    UNKNOWN = -1
    NONE = 0
    UNTRACKED = 1
    ARENA = 2
    BTB = 3
    PVE_BOTS = 4
    FEATURED = 5
    PVE = 6


class SkillResultCode(IntEnum):
    """Result codes for skill requests."""

    SUCCESS = 0
    NOT_FOUND = 1
    SERVICE_FAILURE = 2
    SERVICE_UNAVAILABLE = 3
    FORBIDDEN = 4


class SubTier(Enum):
    """Sub-tiers of skill rankings.

    The `value` attribute of the sub-tier items is the index of the sub-tier,
    starting at 0 for the 1st sub-tier and incrementing as sub-tier increases.
    The `to_int` method returns the true sub-tier value, e.g., "ONE" returns 1.
    """

    I = 0  # noqa: E741
    II = 1
    III = 2
    IV = 3
    V = 4
    VI = 5

    @classmethod
    def from_int(cls, value: int) -> "SubTier":
        """Return the sub-tier from an integer value."""
        return cls(value - 1)

    def to_int(self) -> int:
        """Return the integer value of the sub-tier. For example, "ONE" => 1."""
        return self.value + 1


class Tier(StrEnum):
    """Tiers of skill rankings."""

    UNRANKED = ""
    BRONZE = "Bronze"
    SILVER = "Silver"
    GOLD = "Gold"
    PLATINUM = "Platinum"
    DIAMOND = "Diamond"
    ONYX = "Onyx"



"""Skill data services."""

from pathlib import Path
from typing import Iterable
from uuid import UUID

from spnkr.models.skill import MatchSkill, PlaylistCsr
from spnkr.responses import JsonResponse
from spnkr.services.base import BaseService
from spnkr.xuid import wrap_xuid

_HOST = "https://skill.svc.halowaypoint.com:443"


class SkillService(BaseService):
    """Skill data services."""

    async def get_match_skill(
        self, match_id: str | UUID, xuids: Iterable[str | int]
    ) -> JsonResponse[MatchSkill]:
        """Get player CSR and team MMR values for a given match and player list.

        Args:
            match_id: Halo Infinite match ID.
            xuids: The Xbox Live IDs of the match's players. Only
                players in this list will have their skill data returned.

        Returns:
            The skill data for the match.

        Raises:
            TypeError: If `xuids` is a `str` instead of an iterable of XUIDs.
            ValueError: If `xuids` is empty.
        """
        if isinstance(xuids, str):
            raise TypeError("`xuids` must be an iterable of XUIDs, got `str`")
        if not xuids:
            raise ValueError("`xuids` cannot be empty")
        url = f"{_HOST}/hi/matches/{match_id}/skill"
        params = {"players": [wrap_xuid(x) for x in xuids]}
        resp = await self._get(url, params=params)
        return JsonResponse(resp, lambda data: MatchSkill(**data))

    async def get_playlist_csr(
        self,
        playlist_id: str | UUID,
        xuids: Iterable[str | int],
        season_id: str | None = None,
    ) -> JsonResponse[PlaylistCsr]:
        """Get player CSR values for a given playlist and player list.

        Args:
            playlist_id: Halo Infinite playlist asset ID.
            xuids: The Xbox Live IDs of the players.
            season_id: Halo Infinite season ID. If not provided, the
                current season will be used. Value can be provided as a path,
                e.g. "Csr/Seasons/CsrSeason5-1.json", or as a bare ID,
                e.g. "CsrSeason5-1". If provided as a path, the path prefix
                and ".json" extension will be removed. Case is not important.

        Returns:
            The summary CSR data for the players in the given playlist.

        Raises:
            TypeError: If `xuids` is a `str` instead of an iterable of XUIDs.
            ValueError: If `xuids` is empty.
        """
        if isinstance(xuids, str):
            raise TypeError("`xuids` must be an iterable of XUIDs, got `str`")
        if not xuids:
            raise ValueError("`xuids` cannot be empty")
        url = f"{_HOST}/hi/playlist/{playlist_id}/csrs"
        params: dict = {"players": [wrap_xuid(x) for x in xuids]}
        if season_id:
            params["season"] = _clean_season_id(season_id)
        resp = await self._get(url, params=params)
        return JsonResponse(resp, lambda data: PlaylistCsr(**data))


def _clean_season_id(season_id: str) -> str:
    """Remove the path prefix and ".json" extension from a season ID.

    For example, season "Csr/Seasons/CsrSeason5-1.json" is cleaned to
    "CsrSeason5-1".
    """
    return Path(season_id).stem



import math
from dataclasses import dataclass

from spnkr.models.refdata import SubTier, Tier

BOT_MAP: dict[str, str] = {
    "bid(0.0)": "343 Ritzy",
    "bid(1.0)": "343 Meowlnir",
    "bid(2.0)": "343 Wiggle Cat",
    "bid(3.0)": "343 Ellis",
    "bid(4.0)": "343 Godfather",
    "bid(5.0)": "343 Colson",
    "bid(6.0)": "343 Donos",
    "bid(7.0)": "343 PardonMy",
    "bid(8.0)": "343 Ensrude",
    "bid(9.0)": "343 TooMilks",
    "bid(10.0)": "343 Beard",
    "bid(11.0)": "343 Razzle",
    "bid(12.0)": "343 Nando",
    "bid(13.0)": "343 GrappleMans",
    "bid(14.0)": "343 Zero",
    "bid(15.0)": "343 Mak",
    "bid(16.0)": "343 Hundy",
    "bid(17.0)": "343 Lacuna",
    "bid(18.0)": "343 Cream Corn",
    "bid(19.0)": "343 Brew Dog",
    "bid(20.0)": "343 Lemondade",
    "bid(21.0)": "343 Flippant",
    "bid(22.0)": "343 Connmando",
    "bid(23.0)": "343 Byrontron",
    "bid(24.0)": "343 Ham Sammich",
    "bid(25.0)": "343 BoboGan",
    "bid(26.0)": "343 Robot Hoida",
    "bid(27.0)": "343 Mumblebee",
    "bid(28.0)": "343 Tedosaur",
    "bid(29.0)": "343 Cliffton",
    "bid(30.0)": "343 Stone",
    "bid(31.0)": "343 Marmot",
    "bid(32.0)": "343 Mickey",
    "bid(33.0)": "343 Bachici",
    "bid(34.0)": "343 Ben Desk",
    "bid(35.0)": "343 BF Scrub",
    "bid(36.0)": "343 Chaco",
    "bid(37.0)": "343 Total Ten",
    "bid(38.0)": "343 Darkstar",
    "bid(39.0)": "343 Aloysius",
    "bid(40.0)": "343 Dazzle",
    "bid(41.0)": "343 Rhinosaurus",
    "bid(42.0)": "343 Doomfruit",
    "bid(43.0)": "343 Cosmo",
    "bid(44.0)": "343 Sandwolf",
    "bid(45.0)": "343 O Freruner",
    "bid(46.0)": "343 Bergerton",
    "bid(47.0)": "343 SpaceCase",
    "bid(48.0)": "343 KaleDucky",
    "bid(49.0)": "343 BERRYHILL",
    "bid(50.0)": "343 Oscar",
    "bid(51.0)": "343 The Referee",
    "bid(52.0)": "343 Free Money",
    "bid(53.0)": "343 Kubly",
    "bid(54.0)": "343 Tanuki",
    "bid(55.0)": "343 Shady Seal",
    "bid(56.0)": "343 Chilies",
    "bid(57.0)": "343 The Thumb",
    "bid(58.0)": "343 Forge Lord",
    "bid(59.0)": "343 Hollis",
}
"""Mapping of bot IDs to names. For example, 'bid(1.0)' -> '343 Connmando'."""
TEAM_MAP: dict[int, str] = {
    0: "Eagle",
    1: "Cobra",
    2: "Hades",
    3: "Valkyrie",
    4: "Rampart",
    5: "Cutlass",
    6: "Valor",
    7: "Hazard",
    8: "Observer",
}
"""Mapping of team IDs to names. For example, 0 -> 'Eagle'."""


@dataclass(frozen=True)
class CompetitiveSkillRank:
    """Represents a Competitive Skill Rank (CSR).

    CSR is broken up into a rank/tier and sub-rank/sub-tier. Each tier increment
    (Bronze to Onyx) represents an increase of 300 points and each sub-tier increment
    (1-6) represents an increase of 50 points (Onyx doesn't have sub-tiers).
    Bronze 1 starts at 0 and any value greater than or equal to 1500 is Onyx rank.

    - BRONZE (1-6)
    - SILVER (1-6)
    - GOLD (1-6)
    - PLATINUM (1-6)
    - DIAMOND (1-6)
    - ONYX

    Attributes:
        csr: The CSR value.

    Examples:
        >>> from spnkr.extras import CompetitiveSkillRank
        >>> str(CompetitiveSkillRank(25))
        'Bronze I'
        >>>
        >>> csr = CompetitiveSkillRank(1065)
        >>> (csr.tier, csr.sub_tier)
        (<Tier.PLATINUM: 'Platinum'>, <SubTier.IV: 3>)
    """

    csr: int | float

    TIERS = [
        Tier.BRONZE,
        Tier.SILVER,
        Tier.GOLD,
        Tier.PLATINUM,
        Tier.DIAMOND,
        Tier.ONYX,
    ]

    @property
    def tier(self) -> Tier:
        """The tier of the ranking (e.g., "Diamond")."""
        return self.TIERS[min(math.floor(self.csr) // 300, 5)]

    @property
    def sub_tier(self) -> SubTier:
        """The sub-tier of the ranking (1-6).

        Onyx doesn't have sub-tiers, but a return value of `SubTier.I` in that case
        aligns with the skill payload.
        """
        if self.tier is Tier.ONYX:
            return SubTier.I
        return SubTier((math.floor(self.csr) % 300) // 50)  # SubTier is 0-based

    def __str__(self) -> str:
        if self.tier is Tier.ONYX:
            return self.tier.value
        return f"{self.tier.value} {self.sub_tier.name}"



"""Models for the "skill" authority."""

from uuid import UUID

from pydantic import field_validator

from spnkr.models.base import PascalCaseModel
from spnkr.models.refdata import SkillResultCode, SubTier, Tier
from spnkr.models.types import ReadOnlyDict


class CsrContainer(PascalCaseModel, frozen=True):
    """Container for a player's CSR info.

    Attributes:
        value: The player's CSR value.
        measurement_matches_remaining: The number of matches remaining until the player's CSR is determined.
        tier: The player's CSR tier.
        tier_start: The CSR value at which the player's current tier starts.
        sub_tier: The player's CSR sub-tier.
        next_tier: The player's next CSR tier.
        next_tier_start: The CSR value at which the player's next tier starts.
        next_sub_tier: The player's next CSR sub-tier.
        initial_measurement_matches: The number of matches required to determine the player's CSR.
    """

    value: int
    measurement_matches_remaining: int
    tier: Tier
    tier_start: int
    sub_tier: SubTier
    next_tier: Tier
    next_tier_start: int
    next_sub_tier: SubTier
    initial_measurement_matches: int


class RankRecap(PascalCaseModel, frozen=True):
    """Summary of the player's CSR change.

    Attributes:
        pre_match_csr: The player's CSR details before the match.
        post_match_csr: The player's CSR details after the match.
    """

    pre_match_csr: CsrContainer
    post_match_csr: CsrContainer


class StatPerformance(PascalCaseModel, frozen=True):
    """Comparison of actual to expected values for a game metric.

    Attributes:
        count: The actual value.
        expected: The expected value.
        std_dev: The standard deviation between actual and expected values.
    """

    count: int
    expected: float
    std_dev: float


class StatPerformances(PascalCaseModel, frozen=True):
    """Comparison of actual to expected values for game metrics.

    Attributes:
        kills: Comparison of actual to expected kills.
        deaths: Comparison of actual to expected deaths.
    """

    kills: StatPerformance
    deaths: StatPerformance


class Counterfactual(PascalCaseModel, frozen=True):
    """Expected performance for a given player or skill tier in a match.

    Attributes:
        kills: Expected kills.
        deaths: Expected deaths.
    """

    kills: float
    deaths: float


class Counterfactuals(PascalCaseModel, frozen=True):
    """Expected performances for the player and all skill tiers in a match.

    Attributes:
        self_counterfactuals: Expected performance for the player.
        tier_counterfactuals: Expected performances for all skill tiers.
    """

    self_counterfactuals: Counterfactual
    tier_counterfactuals: ReadOnlyDict[Tier, Counterfactual]


class RankedRewards(PascalCaseModel, frozen=True):
    """Rewards awarded to players based on skill acheivement.

    Attributes:
        reward_id: The ID of the player's reward.
    """

    reward_id: UUID


class MatchSkillResult(PascalCaseModel, frozen=True):
    """Skill data for a player in a match.

    Attributes:
        team_mmr: The MMR of the player's team.
        rank_recap: Summary of the player's CSR change.
        stat_performances: Comparison of actual to expected values for player performance metrics.
            `None` if the skill result could not be found.
        team_id: The ID of the player's team.
        team_mmrs: The MMRs of all teams in the match.
        ranked_rewards: Always null.
        counterfactuals: Expected performances for the player and all skill tiers in a match.
            `None` if the skill result could not be found.
    """

    team_mmr: float
    rank_recap: RankRecap
    stat_performances: StatPerformances | None
    team_id: int
    team_mmrs: ReadOnlyDict[int, float]
    ranked_rewards: RankedRewards | None
    counterfactuals: Counterfactuals | None

    @field_validator("stat_performances", mode="before")
    @classmethod
    def _set_stat_performances_to_none(cls, v):
        """If `stat_performances` is falsy (e.g. {}), set it to None."""
        return v if v else None


class MatchSkillValue(PascalCaseModel, frozen=True):
    """Skill data for a player in a match.

    Attributes:
        id: The player's Xbox Live ID.
        result_code: The status of the skill result.
        result: Skill data for a player in a match.
    """

    id: str
    result_code: SkillResultCode
    result: MatchSkillResult


class MatchSkill(PascalCaseModel, frozen=True):
    """Summary of skill data for teams and players in a match.

    Attributes:
        value: List of player skill data entries.
    """

    value: tuple[MatchSkillValue, ...]


class PlaylistCsrResult(PascalCaseModel, frozen=True):
    """CSR details for a player in a playlist.

    Attributes:
        current: Details on the player's current CSR in the playlist.
        season_max: Details on the player's highest CSR in the playlist for the season.
        all_time_max: Details on the player's highest CSR in the playlist for all seasons.
    """

    current: CsrContainer
    season_max: CsrContainer
    all_time_max: CsrContainer


class PlaylistCsrValue(PascalCaseModel, frozen=True):
    """Playlist CSR result for a single player.

    Attributes:
        id: The player's Xbox Live ID.
        result_code: The status of the skill result.
        result: CSR details for a player in a playlist.
    """

    id: str
    result_code: SkillResultCode
    result: PlaylistCsrResult


class PlaylistCsr(PascalCaseModel, frozen=True):
    """Summary of CSR data for players in a playlist.

    Attributes:
        value: List of playlist-player CSR entries.
    """

    value: tuple[PlaylistCsrValue, ...]



"""Download and read Halo Infinite film chunks to extract match data."""

from spnkr.film.api import read_highlight_events
from spnkr.film.highlight_events import HighlightEvent

__all__ = ["read_highlight_events", "HighlightEvent"]



"""Models for the stats authority."""

import datetime as dt
from uuid import UUID

from pydantic import Field

from spnkr.models.base import PascalCaseModel
from spnkr.models.refdata import (
    AssetKind,
    BotDifficulty,
    GameplayInteraction,
    GameVariantCategory,
    LifecycleMode,
    MedalNameId,
    Outcome,
    PersonalScoreNameId,
    PlayerType,
    PlaylistExperience,
)


class MatchCount(PascalCaseModel, frozen=True):
    """A player's match count summary.

    Attributes:
        custom_matches_played_count: The number of custom matches the player has played.
        matches_played_count: The number of matches the player has played.
        matchmade_matches_played_count: The number of matchmade matches the player has played.
        local_matches_played_count: The number of local matches the player has played.
    """

    custom_matches_played_count: int
    matches_played_count: int
    matchmade_matches_played_count: int
    local_matches_played_count: int


class Asset(PascalCaseModel, frozen=True):
    """ID information about a game asset, such as a map or game mode.

    Attributes:
        asset_kind: The kind of asset, such as "map variant" or "playlist".
        asset_id: The asset's GUID.
        version_id: The asset version's GUID.
    """

    asset_kind: AssetKind
    asset_id: UUID
    version_id: UUID


class MatchInfo(PascalCaseModel, frozen=True):
    """Information about a match.

    Attributes:
        start_time: The UTC datetime when the match started.
        end_time: The UTC datetime when the match ended.
        duration: The match's duration.
        lifecycle_mode: The match's lifecycle mode.
        game_variant_category: The game mode category.
        level_id: The ID of the level the match was played on.
        map_variant: The map the match was played on.
        ugc_game_variant: The game mode the match was played with.
        clearance_id: Clearance ID.
        playlist: The playlist the match was made in, if the match was matchmade.
        playlist_experience: The general category of playlist the match was made in, if the match was matchmade.
        playlist_map_mode_pair: The map and game mode the match was made in, if the match was matchmade.
        season_id: The ID of the season the match was played in, if the match was matchmade.
        playable_duration: The match's playable duration.
        teams_enabled: Whether teams were enabled in the match.
        team_scoring_enabled: Whether team scoring was enabled in the match.
        gameplay_interaction: Types of enemies fought in the match (e.g., PVP).
    """

    start_time: dt.datetime
    end_time: dt.datetime
    duration: dt.timedelta
    lifecycle_mode: LifecycleMode
    game_variant_category: GameVariantCategory
    level_id: UUID
    map_variant: Asset
    ugc_game_variant: Asset
    clearance_id: UUID
    playlist: Asset | None
    playlist_experience: PlaylistExperience | None
    playlist_map_mode_pair: Asset | None
    season_id: str | None
    playable_duration: dt.timedelta
    teams_enabled: bool
    team_scoring_enabled: bool
    gameplay_interaction: GameplayInteraction


class MatchHistoryResult(PascalCaseModel, frozen=True):
    """A match history entry.

    Attributes:
        match_id: The match's GUID.
        match_info: Information about the match.
        last_team_id: The ID of the last team the player was on.
        outcome: The player's outcome in the match.
        rank: The player's rank in the match.
        present_at_end_of_match: Whether the player was present at the end of the match.
    """

    match_id: UUID
    match_info: MatchInfo
    last_team_id: int
    outcome: Outcome
    rank: int
    present_at_end_of_match: bool


class MatchHistory(PascalCaseModel, frozen=True):
    """A page of a player's match history.

    Attributes:
        start: The index of the first match in the page. 0 is the most recent match.
        count: The number of matches requested.
        result_count: The actual number of matches in the page.
        results: The matches in the page.
    """

    start: int
    count: int
    result_count: int
    results: tuple[MatchHistoryResult, ...]


class MatchResultMedalStats(PascalCaseModel, frozen=True):
    """A count of a medal award.

    Attributes:
        name_id: The ID of the medal award.
        count: The number of times the medal award was earned.
        total_personal_score_awarded: The total personal score awarded by obtaining the medal award.
    """

    name_id: MedalNameId
    count: int
    total_personal_score_awarded: int


class MatchResultPersonalScoreStats(PascalCaseModel, frozen=True):
    """A count of a personal score award.

    Attributes:
        name_id: The ID of the personal score award.
        count: The number of times the personal score award was earned.
        total_personal_score_awarded: The total personal score awarded by obtaining the personal score award.
    """

    name_id: PersonalScoreNameId
    count: int
    total_personal_score_awarded: int


class _CoreStats(PascalCaseModel, frozen=True):
    """Core statistics about a player or team performance in a match."""

    score: int
    personal_score: int
    rounds_won: int
    rounds_lost: int
    rounds_tied: int
    kills: int
    deaths: int
    assists: int
    suicides: int
    betrayals: int
    grenade_kills: int
    headshot_kills: int
    melee_kills: int
    power_weapon_kills: int
    shots_fired: int
    shots_hit: int
    accuracy: float
    damage_dealt: int
    damage_taken: int
    callout_assists: int
    vehicle_destroys: int
    driver_assists: int
    hijacks: int
    emp_assists: int
    max_killing_spree: int
    medals: tuple[MatchResultMedalStats, ...]
    personal_scores: tuple[MatchResultPersonalScoreStats, ...]
    spawns: int
    objectives_completed: int


class CoreStats(_CoreStats, frozen=True):
    """Core statistics about a player or team performance in a match.

    Attributes:
        score: Earned points related to the match outcome.
        personal_score: Earned points related to personal performance.
        rounds_won: The number of rounds won.
        rounds_lost: The number of rounds lost.
        rounds_tied: The number of rounds tied.
        kills: The number of kills.
        deaths: The number of deaths.
        assists: The number of assists.
        kda: Kill-death-assist metric (kills + assists / 3 - deaths).
        suicides: The number of suicides.
        betrayals: The number of betrayals.
        average_life_duration: The average time between deaths.
        grenade_kills: The number of kills using grenades.
        headshot_kills: The number of kills with headshots.
        melee_kills: The number of kills with melee or a melee weapon.
        power_weapon_kills: The number of kills with "power" weapons.
        shots_fired: The number of shots fired with weapons.
        shots_hit: The number of shots hit with weapons.
        accuracy: The accuracy of shots fired (shots_fired / shots_hit * 100).
        damage_dealt: The amount of damage dealt.
        damage_taken: The amount of damage taken.
        callout_assists: The number of assists earned by "marking" an enemy.
        vehicle_destroys: The number of vehicles destroyed.
        driver_assists: The number of assists earned by driving a teammate that gets a kill.
        hijacks: The number of times hijacking a vehicle.
        emp_assists: The number of assists earned by EMPing an enemy player or vehicle.
        max_killing_spree: The maximum number of kills without dying.
        medals: Medals earned.
        personal_scores: Personal score awards earned.
        spawns: The number of times spawned.
        objectives_completed: Number of objectives completed.
    """

    kda: float = Field(alias="KDA")
    average_life_duration: dt.timedelta


class ServiceRecordCoreStats(_CoreStats, frozen=True):
    """Core statistics about a player or team performance in a match.

    Attributes:
        score: Earned points related to the match outcome.
        personal_score: Earned points related to personal performance.
        rounds_won: The number of rounds won.
        rounds_lost: The number of rounds lost.
        rounds_tied: The number of rounds tied.
        kills: The number of kills.
        deaths: The number of deaths.
        assists: The number of assists.
        average_kda: Kill-death-assist metric (kills + assists / 3 - deaths).
        suicides: The number of suicides.
        betrayals: The number of betrayals.
        grenade_kills: The number of kills using grenades.
        headshot_kills: The number of kills with headshots.
        melee_kills: The number of kills with melee or a melee weapon.
        power_weapon_kills: The number of kills with "power" weapons.
        shots_fired: The number of shots fired with weapons.
        shots_hit: The number of shots hit with weapons.
        accuracy: The accuracy of shots fired (shots_fired / shots_hit * 100).
        damage_dealt: The amount of damage dealt.
        damage_taken: The amount of damage taken.
        callout_assists: The number of assists earned by "marking" an enemy.
        vehicle_destroys: The number of vehicles destroyed.
        driver_assists: The number of assists earned by driving a teammate that gets a kill.
        hijacks: The number of times hijacking a vehicle.
        emp_assists: The number of assists earned by EMPing an enemy player or vehicle.
        max_killing_spree: The maximum number of kills without dying.
        medals: Medals earned.
        personal_scores: Personal score awards earned.
        spawns: The number of times spawned.
        objectives_completed: Number of objectives completed.
    """

    average_kda: float = Field(alias="AverageKDA")


class BombStats(PascalCaseModel, frozen=True):
    """Performance statistics for bomb game modes.

    Attributes:
        bomb_carriers_killed: Number of times an enemy bomb carrier was killed.
        bomb_defusals: Number of times a bomb was defused.
        bomb_defusers_killed: Number of times an enemy bomb defuser was killed.
        bomb_detonations: Number of times a bomb was detonated.
        bomb_pick_ups: Number of times a bomb was picked up.
        bomb_plants: Number of times a bomb was planted.
        bomb_returns: Number of times a bomb was returned.
        kills_as_bomb_carrier: Number of kills while carrying the bomb.
        time_as_bomb_carrier: Total time spent carrying the bomb.
    """

    bomb_carriers_killed: int
    bomb_defusals: int
    bomb_defusers_killed: int
    bomb_detonations: int
    bomb_pick_ups: int
    bomb_plants: int
    bomb_returns: int
    kills_as_bomb_carrier: int
    time_as_bomb_carrier: dt.timedelta


class CaptureTheFlagStats(PascalCaseModel, frozen=True):
    """Performance statistics for capture the flag game modes.

    Attributes:
        flag_capture_assists: Number of times contributing to a flag capture.
        flag_captures: Number of times the enemy flag was captured.
        flag_carriers_killed: Number of times an enemy flag carrier was killed.
        flag_grabs: Number of times the enemy flag was grabbed.
        flag_returners_killed: Number of times an enemy flag returner was killed.
        flag_returns: Number of times the team flag was returned.
        flag_secures: Number of times the team flag was secured.
        flag_steals: Number of times stealing a flag.
        kills_as_flag_carrier: Number of kills while carrying the flag.
        kills_as_flag_returner: Number of kills while returning the flag.
        time_as_flag_carrier: Total time spent carrying the flag.
    """

    flag_capture_assists: int
    flag_captures: int
    flag_carriers_killed: int
    flag_grabs: int
    flag_returners_killed: int
    flag_returns: int
    flag_secures: int
    flag_steals: int
    kills_as_flag_carrier: int
    kills_as_flag_returner: int
    time_as_flag_carrier: dt.timedelta


class _EliminationStats(PascalCaseModel, frozen=True):
    """Performance statistics for elimination game modes."""

    allies_revived: int
    elimination_assists: int
    eliminations: int
    enemy_revives_denied: int
    executions: int
    kills_as_last_player_standing: int
    last_players_standing_killed: int
    rounds_survived: int
    times_revived_by_ally: int


class EliminationStats(_EliminationStats, frozen=True):
    """Performance statistics for elimination game modes.

    Attributes:
        allies_revived: Number of times reviving a teammate.
        elimination_assists: Number of times contributing to an elimination.
        eliminations: Number of times eliminating an enemy.
        enemy_revives_denied: Number of times denying an enemy revive.
        executions: Number of times executing an enemy.
        kills_as_last_player_standing: Number of kills as the last player standing.
        last_players_standing_killed: Number of times killing the last player standing.
        rounds_survived: Number of rounds survived.
        times_revived_by_ally: Number of times revived by a teammate.
        lives_remaining: Number of lives remaining.
        elimination_order: The order eliminated.
    """

    lives_remaining: int | None
    elimination_order: int


class ServiceRecordEliminationStats(_EliminationStats, frozen=True):
    """Performance statistics for elimination game modes.

    Attributes:
        allies_revived: Number of times reviving a teammate.
        elimination_assists: Number of times contributing to an elimination.
        eliminations: Number of times eliminating an enemy.
        enemy_revives_denied: Number of times denying an enemy revive.
        executions: Number of times executing an enemy.
        kills_as_last_player_standing: Number of kills as the last player standing.
        last_players_standing_killed: Number of times killing the last player standing.
        rounds_survived: Number of rounds survived.
        times_revived_by_ally: Number of times revived by a teammate.
    """


class ExtractionStats(PascalCaseModel, frozen=True):
    """Performance statistics for extraction game modes.

    Attributes:
        extraction_conversions_completed: Number of times converting an extraction zone.
        extraction_conversions_denied: Number of times denying an enemy extraction zone conversion.
        extraction_initiations_completed: Number of times initiating an extraction zone.
        extraction_initiations_denied: Number of times denying an enemy extraction zone initiation.
        successful_extractions: Number of times extracting a zone.
    """

    extraction_conversions_completed: int
    extraction_conversions_denied: int
    extraction_initiations_completed: int
    extraction_initiations_denied: int
    successful_extractions: int


class InfectionStats(PascalCaseModel, frozen=True):
    """Performance statistics for infection game modes.

    Attributes:
        alphas_killed: Number of times infecting an alpha zombie.
        spartans_infected: Number of times infecting a spartan.
        spartans_infected_as_alpha: Number of times infecting a spartan as an alpha zombie.
        kills_as_last_spartan_standing: Number of kills as the last spartan standing.
        last_spartans_standing_infected: Number of times infecting the last spartan standing.
        rounds_as_alpha: Number of rounds as an alpha zombie.
        rounds_as_last_spartan_standing: Number of rounds as the last spartan standing.
        rounds_finished_as_infected: Number of rounds finished as an infected.
        rounds_survived_as_spartan: Number of rounds survived as a spartan.
        rounds_survived_as_last_spartan_standing: Number of rounds survived as the last spartan standing.
        time_as_last_spartan_standing: Total time spent as the last spartan standing.
        infected_killed: Number of times killing an infected.
    """

    alphas_killed: int
    spartans_infected: int
    spartans_infected_as_alpha: int
    kills_as_last_spartan_standing: int
    last_spartans_standing_infected: int
    rounds_as_alpha: int
    rounds_as_last_spartan_standing: int
    rounds_finished_as_infected: int
    rounds_survived_as_spartan: int
    rounds_survived_as_last_spartan_standing: int
    time_as_last_spartan_standing: dt.timedelta
    infected_killed: int


class OddballStats(PascalCaseModel, frozen=True):
    """Performance statistics for oddball game modes.

    Attributes:
        kills_as_skull_carrier: Number of kills while carrying the skull.
        longest_time_as_skull_carrier: Longest time spent carrying the skull.
        skull_carriers_killed: Number of times an enemy skull carrier was killed.
        skull_grabs: Number of times the skull was grabbed.
        time_as_skull_carrier: Total time spent carrying the skull.
        skull_scoring_ticks: Number of ticks the skull was held.
    """

    kills_as_skull_carrier: int
    longest_time_as_skull_carrier: dt.timedelta
    skull_carriers_killed: int
    skull_grabs: int
    time_as_skull_carrier: dt.timedelta
    skull_scoring_ticks: int


class ZonesStats(PascalCaseModel, frozen=True):
    """Performance statistics for zones game modes.

    Attributes:
        stronghold_captures: Number of times capturing a stronghold.
        stronghold_defensive_kills: Number of kills defending a stronghold.
        stronghold_offensive_kills: Number of kills attacking a stronghold.
        stronghold_secures: Number of times securing a stronghold.
        stronghold_occupation_time: Total time spent in a stronghold.
        stronghold_scoring_ticks: Number of ticks a stronghold was held.
    """

    stronghold_captures: int
    stronghold_defensive_kills: int
    stronghold_offensive_kills: int
    stronghold_secures: int
    stronghold_occupation_time: dt.timedelta
    stronghold_scoring_ticks: int


class ServiceRecordZonesStats(PascalCaseModel, frozen=True):
    """Performance statistics for zones game modes.

    Attributes:
        zone_captures: Number of times capturing a zone.
        zone_defensive_kills: Number of kills defending a zone.
        zone_offensive_kills: Number of kills attacking a zone.
        zone_secures: Number of times securing a zone.
        total_zone_occupation_time: Total time spent in a zone.
        zone_scoring_ticks: Number of ticks a zone was held.
    """

    zone_captures: int
    zone_defensive_kills: int
    zone_offensive_kills: int
    zone_secures: int
    total_zone_occupation_time: dt.timedelta
    zone_scoring_ticks: int


class StockpileStats(PascalCaseModel, frozen=True):
    """Performance statistics for stockpile game modes.

    Attributes:
        kills_as_power_seed_carrier: Number of kills while carrying a power seed.
        power_seeds_deposited: Number of power seeds deposited.
        power_seeds_stolen: Number of power seeds stolen.
        power_seed_carriers_killed: Number of times an enemy power seed carrier was killed.
        time_as_power_seed_carrier: Total time spent carrying a power seed.
        time_as_power_seed_driver: Total time spent driving a power seed.
    """

    kills_as_power_seed_carrier: int
    power_seeds_deposited: int
    power_seeds_stolen: int
    power_seed_carriers_killed: int
    time_as_power_seed_carrier: dt.timedelta
    time_as_power_seed_driver: dt.timedelta


class _PvpStats(PascalCaseModel, frozen=True):
    """Player vs. player stats."""

    kills: int
    deaths: int
    assists: int


class PvpStats(_PvpStats, frozen=True):
    """Player vs. player stats."""

    kda: float = Field(alias="KDA")


class ServiceRecordPvpStats(_PvpStats, frozen=True):
    """Player vs. player stats."""


class _PveStats(PascalCaseModel, frozen=True):
    """Player vs. environment stats."""

    kills: int
    deaths: int
    assists: int
    marine_kills: int
    grunt_kills: int
    jackal_kills: int
    elite_kills: int
    brute_kills: int
    hunter_kills: int
    skimmer_kills: int
    sentinel_kills: int
    boss_kills: int


class PveStats(_PveStats, frozen=True):
    """Player vs. environment stats."""

    kda: float = Field(alias="KDA")


class ServiceRecordPveStats(_PveStats, frozen=True):
    """Player vs. environment stats."""


class VipStats(PascalCaseModel, frozen=True):
    kills_as_vip: int
    longest_time_as_vip: dt.timedelta
    max_killing_spree_as_vip: int
    time_as_vip: dt.timedelta
    times_selected_as_vip: int
    vip_assists: int
    vip_kills: int


class Stats(PascalCaseModel, frozen=True):
    """Performance statistics for a player or team in a match.

    Attributes:
        core_stats: Performance statistics common to all game modes.
        bomb_stats: Performance statistics for bomb game modes.
        capture_the_flag_stats: Performance statistics for capture the flag game modes.
        elimination_stats: Performance statistics for elimination game modes.
        extraction_stats: Performance statistics for extraction game modes.
        infection_stats: Performance statistics for infection game modes.
        oddball_stats: Performance statistics for oddball game modes.
        zones_stats: Performance statistics for zones game modes.
        stockpile_stats: Performance statistics for stockpile game modes.
        pvp_stats: Player vs. player stats.
        pve_stats: Player vs. environment stats.
    """

    core_stats: CoreStats
    bomb_stats: BombStats | None = None
    capture_the_flag_stats: CaptureTheFlagStats | None = None
    elimination_stats: EliminationStats | None = None
    extraction_stats: ExtractionStats | None = None
    infection_stats: InfectionStats | None = None
    oddball_stats: OddballStats | None = None
    zones_stats: ZonesStats | None = None
    stockpile_stats: StockpileStats | None = None
    pvp_stats: PvpStats | None = None
    pve_stats: PveStats | None = None
    vip_stats: VipStats | None = None


class TeamStats(PascalCaseModel, frozen=True):
    """Statistics for a team in a match.

    Attributes:
        team_id: The team's ID.
        outcome: The team's outcome in the match.
        rank: The team's rank in the match.
        stats: The team's performance statistics.
    """

    team_id: int
    outcome: Outcome
    rank: int
    stats: Stats


class BotAttributes(PascalCaseModel, frozen=True):
    """Attributes of a bot player.

    Attributes:
        difficulty: The bot's difficulty.
    """

    difficulty: BotDifficulty


class ParticipationInfo(PascalCaseModel, frozen=True):
    """Information about a player's participation in a match.

    Attributes:
        first_joined_time: The UTC datetime when the player first joined the match.
        last_leave_time: The UTC datetime when the player last left the match, if the player left the match.
        present_at_beginning: Whether the player was present at the beginning of the match.
        joined_in_progress: Whether the player joined the match in progress.
        left_in_progress: Whether the player left the match in progress.
        present_at_completion: Whether the player was present at the completion of the match.
        time_played: The amount of time the player was present in the match.
        confirmed_participation: Whether the player's participation was confirmed.
    """

    first_joined_time: dt.datetime
    last_leave_time: dt.datetime | None
    present_at_beginning: bool
    joined_in_progress: bool
    left_in_progress: bool
    present_at_completion: bool
    time_played: dt.timedelta
    confirmed_participation: bool | None


class PlayerTeamStats(PascalCaseModel, frozen=True):
    """Statistics for a player on a team in a match.

    Attributes:
        team_id: The team's ID.
        stats: The player's performance statistics while on the team.
    """

    team_id: int
    stats: Stats


class PlayerStats(PascalCaseModel, frozen=True):
    """Statistics for a player in a match.

    Attributes:
        player_id: The player's Xbox Live ID.
        player_type: The player's type, such as "human" or "bot".
        bot_attributes: Details about bot players, if applicable.
        last_team_id: The ID of the last team the player was on.
        outcome: The player's outcome in the match.
        rank: The player's rank in the match.
        participation_info: Information about the player's participation in the match.
        player_team_stats: The player's performance statistics for each team they were on during the match.
    """

    player_id: str
    player_type: PlayerType
    bot_attributes: BotAttributes | None
    last_team_id: int
    outcome: Outcome
    rank: int
    participation_info: ParticipationInfo
    player_team_stats: tuple[PlayerTeamStats, ...]

    @property
    def is_human(self) -> bool:
        """Whether the player is a human player."""
        return self.player_type is PlayerType.HUMAN


class MatchStats(PascalCaseModel, frozen=True):
    """Player and team performance details for a match.

    Attributes:
        match_id: The match's GUID.
        match_info: Information about the match.
        teams: Performance statistics for all teams in the match.
        players: Performance statistics for all players in the match.
    """

    match_id: UUID
    match_info: MatchInfo
    teams: tuple[TeamStats, ...]
    players: tuple[PlayerStats, ...]

    @property
    def xuids(self) -> list[int]:
        """Xbox user IDs for all human players in the match."""
        return [int(p.player_id[5:-1]) for p in self.players if p.is_human]


class ServiceRecordSubqueries(PascalCaseModel, frozen=True):
    """Subqueries for a service record request.

    Attributes:
        season_ids: Seasons available for filtering.
        game_variant_categories: Game variant categories available for filtering.
        is_ranked: Ranked status available for filtering.
        playlist_asset_ids: Playlists available for filtering.
        gameplay_interactions: Gameplay interactions available for filtering.
    """

    season_ids: tuple[str, ...] | None
    game_variant_categories: tuple[GameVariantCategory, ...] | None
    is_ranked: tuple[bool, ...] | None
    playlist_asset_ids: tuple[UUID, ...] | None
    gameplay_interactions: tuple[GameplayInteraction, ...] | None


class ServiceRecord(PascalCaseModel, frozen=True):
    """A player's service record within a given context.

    Attributes:
        subqueries: Subquerying options for the service record request.
        time_played: The player's total time played.
        matches_completed: The number of matches the player has completed.
        wins: The number of matches the player has won.
        losses: The number of matches the player has lost.
        ties: The number of matches the player has tied.
        core_stats: The player's core performance statistics.
        bomb_stats: The player's performance statistics for bomb game modes.
        capture_the_flag_stats: The player's performance statistics for capture
            the flag game modes.
        elimination_stats: The player's performance statistics for elimination
            game modes.
        infection_stats: The player's performance statistics for infection game
            modes.
        oddball_stats: The player's performance statistics for oddball game
            modes.
        zones_stats: The player's performance statistics for zones game modes.
        stockpile_stats: The player's performance statistics for stockpile game
            modes.
        pvp_stats: Player vs. player stats.
        pve_stats: Player vs. environment stats.
    """

    subqueries: ServiceRecordSubqueries
    time_played: dt.timedelta
    matches_completed: int
    wins: int
    losses: int
    ties: int
    core_stats: ServiceRecordCoreStats
    bomb_stats: BombStats | None = None
    capture_the_flag_stats: CaptureTheFlagStats | None = None
    elimination_stats: ServiceRecordEliminationStats | None = None
    infection_stats: InfectionStats | None = None
    oddball_stats: OddballStats | None = None
    zones_stats: ServiceRecordZonesStats | None = None
    stockpile_stats: StockpileStats | None = None
    pvp_stats: ServiceRecordPvpStats | None = None
    pve_stats: ServiceRecordPveStats | None = None



import json
from importlib import resources

_bytes = (resources.files("spnkr.film") / "medal_codes.json").read_bytes()
MEDALS: dict[int, str] = {int(k): v for k, v in json.loads(_bytes).items()}
"""Mapping of int encodings to their respective medal names."""



"""Python API for requesting Halo Infinite multiplayer data"""

from spnkr.auth import AzureApp, authenticate_player, refresh_player_tokens
from spnkr.client import HaloInfiniteClient

__all__ = [
    "AzureApp",
    "authenticate_player",
    "refresh_player_tokens",
    "HaloInfiniteClient",
]



"""Economy/customization data services."""

from typing import Literal

from spnkr.models.economy import PlayerCustomization
from spnkr.responses import JsonResponse
from spnkr.services.base import BaseService
from spnkr.xuid import wrap_xuid

_HOST = "https://economy.svc.halowaypoint.com:443"


class EconomyService(BaseService):
    """Economy/customization data services."""

    async def get_player_customization(
        self,
        xuid: str | int,
        view_type: Literal["public", "private"] = "public",
    ) -> JsonResponse[PlayerCustomization]:
        """Get the customization selections for a player.

        Args:
            xuid: The Xbox Live ID of the player to retrieve data for.
            view_type: Use "public" when requesting customization for players
                other than yourself. The "private" view type only applies to the
                authenticated player making the request and might include data
                not visible using the "public" view type.

        Returns:
            The player's customization selections.
        """
        url = f"{_HOST}/hi/players/{wrap_xuid(xuid)}/customization"
        params = {"view": view_type}
        resp = await self._get(url, params=params)
        return JsonResponse(resp, lambda data: PlayerCustomization(**data))



"""Wrappers for API responses."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar

if TYPE_CHECKING:
    from aiohttp import ClientResponse
    from aiohttp_client_cache.response import CachedResponse

T = TypeVar("T")


@dataclass(frozen=True)
class _BaseResponse:
    response: "ClientResponse | CachedResponse"

    @property
    def from_cache(self) -> bool:
        """Whether the response is from cache or not."""
        return hasattr(self.response, "from_cache")

    async def read(self) -> bytes:
        """Read the response content."""
        return await self.response.read()


@dataclass(frozen=True)
class JsonResponse(_BaseResponse, Generic[T]):
    """`ClientResponse` wrapper for a JSON HTTP response."""

    _parser: Callable[[Any], T]

    async def json(self, **kwargs) -> Any:
        """Read the response body as JSON.

        Keyword arguments are passed to [aiohttp.ClientResponse.json](https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientResponse.json).
        """
        return await self.response.json(**kwargs)

    async def parse(self, **kwargs) -> T:
        """Parse the response data into the appropriate response model.

        Keyword arguments are passed to [aiohttp.ClientResponse.json](https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.ClientResponse.json).
        """
        return self._parser(await self.json(**kwargs))


class ImageResponse(_BaseResponse):
    """`ClientResponse` wrapper for an image HTTP response."""
